# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Provides the inversion of control design pattern but does not contain the fully
    functional container (we can just only create a class that will set various references).

    Once the objects of a container are configured, if they implement the IReferencable interface,
    they are passed a set of references for recreating links between objects in the container.
    If objects implement the IOpenable interface, the :func:`open()` method is called and they start to work.
    Connections to various services are made, after which the objects start,
    the container starts running, and the objects carry out their tasks. When the container starts to close,
    the objects that implement the ICloseable interface are closed via their
    :func:`close()` method (which should make them stop working and disconnect from other services),
    after which objects that implement the IUnreferenceable interface delete various links between objects,
    and, finally, the contains destroys all objects and turns off.

    :class:`BuildReferencesDecorator`, :class:`LinkReferencesDecorator`, and
    :class:`RunReferencesDecorator` - ReferenceDecorators are used during the corresponding building, linking,
    and running stages and are united in :class:`ManagedReferences`, which are extended by :class:`ContainerReferences`.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = ['ReferencesDecorator', 'RunReferencesDecorator', 'LinkReferencesDecorator',
           'BuildReferencesDecorator', 'ManagedReferences', 'ContainerReferences']

from .BuildReferencesDecorator import BuildReferencesDecorator
from .ContainerReferences import ContainerReferences
from .LinkReferencesDecorator import LinkReferencesDecorator
from .ManagedReferences import ManagedReferences
from .ReferencesDecorator import ReferencesDecorator
from .RunReferencesDecorator import RunReferencesDecorator



# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.ReferencesDecorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    References decorator implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Any, List, Optional

from pip_services3_commons.refer import IReferences


class ReferencesDecorator(IReferences):
    """
    Chainable decorator for :class:`IReferences <pip_services3_commons.refer.IReferences>` that allows to inject additional capabilities
    such as automatic component creation, automatic registration and opening.
    """

    def get_one_before(self, reference, locator):
        pass

    def __init__(self, next_references: Optional[IReferences], top_references: Optional[IReferences]):
        """
        Creates a new instance of the decorator.

        :param next_references: the next references or decorator in the chain.

        :param top_references: the decorator at the top of the chain.
        """
        # The next references or decorator in the chain.
        self.base_references: IReferences = next_references if next_references is not None else top_references
        # The decorator at the top of the chain.
        self.parent_references: IReferences = top_references if top_references is not None else next_references

    def put(self, locator: Any = None, component: Any = None):
        """
        Puts a new reference into this reference map.

        :param locator: a locator to find the reference by.
        :param component: a component reference to be added.
        """
        self.base_references.put(locator, component)

    def remove(self, locator: Any) -> Any:
        """
        Removes a previously added component that matches specified locator.
        If many references match the locator, it removes only the first one.
        When all references shall be removed, use :func:`remove_all` method instead.

        :param locator: a locator to remove component

        :return: the removed component component.
        """
        return self.base_references.remove(locator)

    def remove_all(self, locator: Any) -> List[Any]:
        """
        Removes all component references that match the specified locator.

        :param locator: the locator to remove references by.

        :return: a list, containing all removed references.
        """
        return self.base_references.remove_all(locator)

    def get_all_locators(self) -> List[Any]:
        """
        Gets locators for all registered component references in this component map.

        :return: a list with component locators.
        """
        return self.base_references.get_all_locators()

    def get_all(self) -> List[Any]:
        """
        Gets all component references registered in this component map.

        :return: a list with component references.
        """
        return self.base_references.get_all()

    def get_one_optional(self, locator: Any) -> Any:
        """
        Gets an optional component component that matches specified locator.

        :param locator: the locator to find references by.

        :return: a matching component component or null if nothing was found.
        """
        try:
            components = self.find(locator, False)
            return components[0] if len(components) > 0 else None
        except Exception as ex:
            return None

    def get_one_required(self, locator: Any) -> Any:
        """
        Gets a required component component that matches specified locator.

        :param locator: the locator to find a component by.

        :return: a matching component component.

        :raises: a :class:`ReferenceException <pip_services3_commons.refer.ReferenceException.ReferenceException>` when no references found.
        """
        components = self.find(locator, True)
        return components[0] if len(components) > 0 else None

    def get_optional(self, locator: Any) -> List[Any]:
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find references by.

        :return: a list with matching component references or empty list if nothing was found.
        """
        try:
            return self.find(locator, False)
        except Exception as ex:
            return []

    def get_required(self, locator: Any) -> List[Any]:
        """
        Gets all component references that match specified locator.
        At least one component component must be present. If it doesn't the method throws an error.

        :param locator: the locator to find references by.

        :return: a list with matching component references.

        :raises: a :class:`ReferenceException <pip_services3_commons.refer.ReferenceException.ReferenceException>` when no references found.
        """
        return self.find(locator, True)

    def find(self, locator: Any, required: bool) -> List[Any]:
        """
        Gets all component references that match specified locator.

        :param locator: the locator to find a component by.

        :param required: forces to raise an exception if no component is found.

        :return: a list with matching component references.

        :raises: a :class:`ReferenceException <pip_services3_commons.refer.ReferenceException.ReferenceException>` when required is set to true but no references found.
        """
        return self.base_references.find(locator, required)



# -*- coding: utf-8 -*-
"""
    pip_services3_container.__init__
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Contains implementation of the inversion of control container,
    which creates objects and controls their lifecycle(*) using various configurations.

    Using generic containers, we can create more specialized containers – one of which is the process
    container. It represents a system process, receives its configuration file via the command line,
    and creates a container, starts it, reads its configuration, recreates objects, runs them, and then,
    after pressing ctrl-c, turns off and destroys the objects.

    Another example of containers are lambda functions, service fabric containers, and so on.

    ### (*) Compont lifecycle: ###

    External configurations (stored as YAML or JSON) are passed to the container
    and define the structure of objects that need to be recreated in the container.
    Objects can be defined in two ways:
        - using descriptors (using which registered factories can recreate the object)
        - using hard-coded types (objects are recreated directly, based on their type, bypassing factories).

    In addition, various configurations are stored for each object. The container recreates the
    objects and, if they implement the IConfigurable interface, passes them their configurations.
    Once the objects of a container are configured, if they implement the IReferencable interface,
    they are passed a set of references for recreating links between objects in the container.
    If objects implement the IOpenable interface, the :func:`open()` method is called and they
    start to work. Connections to various services are made, after which the objects start, the
    container starts running, and the objects carry out their tasks. When the container
    starts to close, the objects that implement the ICloseable interface are closed via their
    :func:`close()` method (which should make them stop working and disconnect from other services),
    after which objects that implement the IUnreferencable interface delete various links between
    objects, and, finally, the contains destroys all objects and turns off.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

__all__ = ['Container', 'ProcessContainer']

from .Container import Container
from .ProcessContainer import ProcessContainer



# -*- coding: utf-8 -*-
"""
    pip_services3_container.ProcessContainer
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Process container implementation.

    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""

import os
import signal
import sys
import threading
from typing import Optional

from pip_services3_commons.config import ConfigParams
from pip_services3_components.log import ConsoleLogger

from .Container import Container


class ProcessContainer(Container):
    """
    Inversion of control (IoC) container that runs as a system process.
    It processes command line arguments and handles unhandled exceptions and Ctrl-C signal
    to gracefully shutdown the container.

    ### Command line arguments ###
        - --config -c             path to JSON or YAML file with container configuration (default: "./config/config.yml")
        - --param --params -p   value(s) to parameterize the container configuration
        - --help -h               prints the container usage help

        Example:

        .. code-block:: python
        
            container = ProcessContainer()
            container.add_factory(MyComponentFactory())

            container.run()
    """

    __exit_event: threading.Event = None

    def __init__(self, name: str = None, description: str = None):
        """
        Creates a new instance of the container.

        :param name: (optional) a container name (accessible via ContextInfo)

        :param description: (optional) a container description (accessible via ContextInfo)
        """
        super(ProcessContainer, self).__init__(name, description)
        self._config_path = './config/config.yml'
        self._logger = ConsoleLogger()
        self.__exit_event = threading.Event()

    def __get_config_path(self) -> str:
        args = sys.argv
        index = 0
        while index < len(args):
            arg = args[index]
            next_arg = args[index + 1] if index < len(args) - 1 else None
            next_arg = None if next_arg is not None and next_arg.startswith('-') else next_arg
            if next_arg is not None:
                if arg == "--config" or arg == "-c":
                    return next_arg
            index += 1
        return self._config_path

    def __get_parameters(self) -> ConfigParams:
        # Process command line parameters
        args = sys.argv
        line = ''
        index = 0
        while index < len(args):
            arg = args[index]
            next_arg = args[index + 1] if index < len(args) - 1 else None
            next_arg = None if next_arg is not None and next_arg.startswith('-') else next_arg
            if next_arg is not None:
                if arg == "--param" or arg == "--params" or arg == "-p":
                    if len(line) > 0:
                        line = line + ';'
                    line = line + next_arg
                    index += 1
            index += 1

        parameters = ConfigParams.from_string(line)

        # Process environmental variables
        for (k, v) in os.environ.items():
            parameters[k] = v

        return parameters

    def __show_help(self) -> bool:
        args = sys.argv
        index = 0
        while index < len(args):
            arg = args[index]
            if arg == "--help" or arg == "-h":
                return True
            index += 1
        return False

    def __print_help(self):
        print("Pip.Services process container - http://www.pipservices.org")
        print("run [-h] [-c <config file>] [-p <param>=<value>]*")

    def __capture_errors(self, correlation_id: Optional[str]):
        def handle_exception(exc_type, exc_value, exc_traceback):
            self._logger.fatal(correlation_id, exc_value, "Process is terminated")
            self.__exit_event.set()
            # sys.exit(1)

        sys.excepthook = handle_exception

    def __capture_exit(self, correlation_id: Optional[str]):
        self._logger.info(correlation_id, "Press Control-C to stop the microservice...")

        def sigint_handler(signum, frame):
            self._logger.info(correlation_id, "Goodbye!")
            self.__exit_event.set()
            # sys.exit(1)

        signal.signal(signal.SIGINT, sigint_handler)
        signal.signal(signal.SIGTERM, sigint_handler)

        # Wait and close
        self.__exit_event.clear()
        while not self.__exit_event.is_set():
            try:
                self.__exit_event.wait(1)
            except:
                pass  # Do nothing...

    def run(self):
        """
        Runs the container by instantiating and running components inside the container.

        It reads the container configuration, creates, configures, references and opens components.
        On process exit it closes, unreferences and destroys components to gracefully shutdown.
        """
        if self.__show_help():
            self.__print_help()
            return

        correlation_id = self._info.name
        path = self.__get_config_path()
        parameters = self.__get_parameters()
        self.read_config_from_file(correlation_id, path, parameters)

        self.__capture_errors(correlation_id)
        self.open(correlation_id)
        self.__capture_exit(correlation_id)
        self.close(correlation_id)



# -*- coding: utf-8 -*-
"""
    pip_services3_container.config.ContainerConfig
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Container configuration implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Sequence, Any

from pip_services3_commons.config import ConfigParams

from .ComponentConfig import ComponentConfig


class ContainerConfig(list):
    """
    Container configuration defined as a list of component configurations.
    """

    def __init__(self, components: Sequence[ComponentConfig] = None):
        """
        Creates a new instance of container configuration.

        :param components: (optional) a list of component configurations.
        """
        super(ContainerConfig, self).__init__()
        if components is not None:
            for component in components:
                self.append(component)

    @staticmethod
    def from_value(value: Any) -> 'ContainerConfig':
        """
        Creates a new :class:`ContainerConfig <pip_services3_container.config.ContainerConfig.ContainerConfig>` object filled with key-value pairs from specified object.
        The value is converted into :class:`ConfigParams <pip_services3_commons.config.ConfigParams.ConfigParams>` object which is used to create the object.

        :param value: an object with key-value pairs used to initialize a new :class:`ContainerConfig <pip_services3_container.config.ContainerConfig.ContainerConfig>`.

        :return: a new :class:`ContainerConfig <pip_services3_container.config.ContainerConfig.ContainerConfig>` object.
        """
        config = ConfigParams.from_value(value)
        return ContainerConfig.from_config(config)

    @staticmethod
    def from_config(config: ConfigParams) -> 'ContainerConfig':
        """
        Creates a new :class:`ContainerConfig <pip_services3_container.config.ContainerConfig.ContainerConfig>` object based on configuration parameters.
        Each section in the configuration parameters is converted into a component configuration.

        :param config: an object with key-value pairs used to initialize a new :class:`ContainerConfig <pip_services3_container.config.ContainerConfig.ContainerConfig>`.

        :return: a new :class:`ContainerConfig <pip_services3_container.config.ContainerConfig.ContainerConfig>` object.
        """
        result = ContainerConfig()
        if config is None:
            return result

        for section in config.get_section_names():
            component_config = config.get_section(section)
            result.append(ComponentConfig.from_config(component_config))

        return result



# -*- coding: utf-8 -*-
"""
    pip_services3_container.config.ComponentConfig
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Component configuration implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from pip_services3_commons.config import ConfigParams
from pip_services3_commons.errors import ConfigException
from pip_services3_commons.refer import Descriptor
from pip_services3_commons.reflect import TypeDescriptor


class ComponentConfig:
    """
    Configuration of a component inside a container.

    The configuration includes type information or descriptor, and component configuration parameters.
    """

    def __init__(self, descriptor: Descriptor = None, type: TypeDescriptor = None, config: ConfigParams = None):
        """
        Creates a new instance of the component configuration.

        :param descriptor: (optional) a components descriptor (locator).

        :param type: (optional) a components type descriptor.

        :param config: (optional) component configuration parameters.
        """
        self.descriptor: Descriptor = descriptor
        self.type: TypeDescriptor = type
        self.config: ConfigParams = config

    @staticmethod
    def from_config(config: ConfigParams) -> 'ComponentConfig':
        """
        Creates a new instance of ComponentConfig based on section from container configuration.

        :param config: component parameters from container configuration

        :return: a newly created ComponentConfig
        """
        descriptor = Descriptor.from_string(config.get_as_nullable_string("descriptor"))
        type = TypeDescriptor.from_string(config.get_as_nullable_string("type"))

        if descriptor is None and type is None:
            raise ConfigException(None, "BAD_CONFIG", "Component configuration must have descriptor or type")

        return ComponentConfig(descriptor, type, config)



# -*- coding: utf-8 -*-
"""
    pip_services3_container.refer.ManagedReferences
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    Managed references implementation.
    
    :copyright: Conceptual Vision Consulting LLC 2018-2019, see AUTHORS for more details.
    :license: MIT, see LICENSE for more details.
"""
from typing import Sequence, Any, Optional

from pip_services3_commons.refer import References
from pip_services3_commons.run import IOpenable

from .BuildReferencesDecorator import BuildReferencesDecorator
from .LinkReferencesDecorator import LinkReferencesDecorator
from .ReferencesDecorator import ReferencesDecorator
from .RunReferencesDecorator import RunReferencesDecorator


class ManagedReferences(ReferencesDecorator, IOpenable):
    """
    Managed references that in addition to keeping and locating references can also manage their lifecycle:
        - Auto-creation of missing component using available factories
        - Auto-linking newly added components
        - Auto-opening newly added components
        - Auto-closing removed components
    """

    def __init__(self, tuples: Sequence[Any] = None):
        """
        Creates a new instance of the references

        :param tuples: tuples where odd values are component locators (descriptors) and even values are component references
        """
        super(ManagedReferences, self).__init__(None, None)

        self._references: References = References(tuples)
        self._builder: BuildReferencesDecorator = BuildReferencesDecorator(self._references, self)
        self._linker: LinkReferencesDecorator = LinkReferencesDecorator(self._builder, self)
        self._runner: RunReferencesDecorator = RunReferencesDecorator(self._linker, self)

        self.base_references = self._runner

    def is_open(self) -> bool:
        """
        Checks if the component is opened.

        :return: true if the component has been opened and false otherwise.
        """
        return self._linker.is_open() and self._runner.is_open()

    def open(self, correlation_id: Optional[str]):
        """
        Opens the component.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        self._linker.open(correlation_id)
        self._runner.open(correlation_id)

    def close(self, correlation_id: Optional[str]):
        """
        Closes component and frees used resources.

        :param correlation_id: (optional) transaction id to trace execution through call chain.
        """
        self._runner.close(correlation_id)
        self._linker.close(correlation_id)

    @staticmethod
    def from_tuples(*tuples: Any) -> 'ManagedReferences':
        """
        Creates a new ManagedReferences object filled with provided key-value pairs called tuples.
        Tuples parameters contain a sequence of locator1, component1, locator2, component2, ... pairs.

        :param tuples: the tuples to fill a new ManagedReferences object.

        :return: a new ManagedReferences object.
        """
        return ManagedReferences(tuples)
